@@ -1,5 +1,5 @@
 /* csplit - split a file into sections determined by context lines
-   Copyright (C) 91, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -21,49 +21,34 @@
 #include <config.h>
 
 #include <stdio.h>
-
-/* Disable assertions.  Some systems have broken assert macros.  */
-#define NDEBUG 1
-
 #include <assert.h>
 #include <getopt.h>
 #include <sys/types.h>
 #include <signal.h>
-#ifdef HAVE_LIMITS_H
-#include <limits.h>
-#endif /* HAVE_LIMITS_H */
 
-#ifndef UINT_MAX
-# define UINT_MAX ((unsigned int) ~(unsigned int) 0)
-#endif
-
-#ifndef INT_MAX
-# define INT_MAX ((int) (UINT_MAX >> 1))
-#endif
-
-#if WITH_REGEX
-# include <regex.h>
-#else
-# include <rx.h>
-#endif
 #include "system.h"
+#include "closeout.h"
+
+#include <regex.h>
+
 #include "error.h"
-#include "xstrtoul.h"
+#include "human.h"
+#include "safe-read.h"
+#include "xstrtol.h"
+#include "xalloc.h"
+
+/* The official name of this program (e.g., no `g' prefix).  */
+#define PROGRAM_NAME "csplit"
+
+#define AUTHORS "Stuart Kemp and David MacKenzie"
 
 #ifdef STDC_HEADERS
-#include <stdlib.h>
-#else
-char *malloc ();
-char *realloc ();
-#endif
-
-#ifndef MAX
-#define MAX(a,b) (((a) > (b)) ? (a) : (b))
+# include <stdlib.h>
 #endif
 
 #ifndef TRUE
-#define FALSE 0
-#define TRUE 1
+# define FALSE 0
+# define TRUE 1
 #endif
 
 /* Increment size of area for control records. */
@@ -80,8 +65,8 @@ struct control
   char *regexpr;		/* Non-compiled regular expression. */
   struct re_pattern_buffer re_compiled;	/* Compiled regular expression. */
   int offset;			/* Offset from regexp to split at. */
-  int lines_required;		/* Number of lines required. */
-  unsigned int repeat;		/* Repeat count. */
+  uintmax_t lines_required;	/* Number of lines required. */
+  uintmax_t repeat;		/* Repeat count. */
   int repeat_forever;		/* Non-zero if `*' used as a repeat count. */
   int argnum;			/* ARGV index. */
   boolean ignore;		/* If true, produce no output (for regexp). */
@@ -98,9 +83,9 @@ struct control
 
 #ifdef DEBUG
 /* Some small values to test the algorithms. */
-#define START_SIZE	200
-#define INCR_SIZE	10
-#define CTRL_SIZE	1
+# define START_SIZE	200
+# define INCR_SIZE	10
+# define CTRL_SIZE	1
 #endif
 
 /* A string with a length count. */
@@ -137,13 +122,11 @@ struct buffer_record
   struct buffer_record *next;
 };
 
-int safe_read ();
-
-static void close_output_file __P ((void));
-static void create_output_file __P ((void));
-static void delete_all_files __P ((void));
-static void save_line_to_file __P ((const struct cstring *line));
-static void usage __P ((int status));
+static void close_output_file PARAMS ((void));
+static void create_output_file PARAMS ((void));
+static void delete_all_files PARAMS ((void));
+static void save_line_to_file PARAMS ((const struct cstring *line));
+void usage PARAMS ((int status));
 
 /* The name this program was run with. */
 char *program_name;
@@ -222,12 +205,6 @@ static struct control *controls;
 /* Number of elements in `controls'. */
 static unsigned int control_used;
 
-/* If nonzero, display usage information and exit.  */
-static int show_help;
-
-/* If nonzero, print the version on standard output then exit.  */
-static int show_version;
-
 static struct option const longopts[] =
 {
   {"digits", required_argument, NULL, 'n'},
@@ -237,8 +214,8 @@ static struct option const longopts[] =
   {"elide-empty-files", no_argument, NULL, 'z'},
   {"prefix", required_argument, NULL, 'f'},
   {"suffix-format", required_argument, NULL, 'b'},
-  {"help", no_argument, &show_help, 1},
-  {"version", no_argument, &show_version, 1},
+  {GETOPT_HELP_OPTION_DECL},
+  {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
@@ -265,60 +242,20 @@ cleanup_fatal (void)
 static RETSIGTYPE
 interrupt_handler (int sig)
 {
-#ifdef SA_INTERRUPT
+#ifdef SA_NOCLDSTOP
   struct sigaction sigact;
 
   sigact.sa_handler = SIG_DFL;
   sigemptyset (&sigact.sa_mask);
   sigact.sa_flags = 0;
   sigaction (sig, &sigact, NULL);
-#else				/* !SA_INTERRUPT */
+#else
   signal (sig, SIG_DFL);
-#endif				/* SA_INTERRUPT */
+#endif
   cleanup ();
   kill (getpid (), sig);
 }
 
-/* Allocate N bytes of memory dynamically, with error checking.  */
-
-static char *
-xmalloc (unsigned int n)
-{
-  char *p;
-
-  p = malloc (n);
-  if (p == NULL)
-    {
-      error (0, 0, _("virtual memory exhausted"));
-      cleanup_fatal ();
-    }
-  return p;
-}
-
-/* Change the size of an allocated block of memory P to N bytes,
-   with error checking.
-   If P is NULL, run xmalloc.
-   If N is 0, run free and return NULL.  */
-
-static char *
-xrealloc (char *p, unsigned int n)
-{
-  if (p == NULL)
-    return xmalloc (n);
-  if (n == 0)
-    {
-      free (p);
-      return 0;
-    }
-  p = realloc (p, n);
-  if (p == NULL)
-    {
-      error (0, 0, _("virtual memory exhausted"));
-      cleanup_fatal ();
-    }
-  return p;
-}
-
 /* Keep track of NUM chars of a partial line in buffer START.
    These chars will be retrieved later when another large buffer is read.
    It is not necessary to create a new buffer for these chars; instead,
@@ -741,7 +678,7 @@ no_more_lines (void)
 static void
 set_input_file (const char *name)
 {
-  if (!strcmp (name, "-"))
+  if (STREQ (name, "-"))
     input_desc = 0;
   else
     {
@@ -804,8 +741,10 @@ dump_rest_of_file (void)
 static void
 handle_line_error (const struct control *p, int repetition)
 {
-  fprintf (stderr, _("%s: `%d': line number out of range"),
-	   program_name, p->lines_required);
+  char buf[LONGEST_HUMAN_READABLE + 1];
+
+  fprintf (stderr, _("%s: `%s': line number out of range"),
+	   program_name, human_readable (p->lines_required, buf, 1, 1));
   if (repetition)
     fprintf (stderr, _(" on repetition %d\n"), repetition);
   else
@@ -823,7 +762,7 @@ static void
 process_line_count (const struct control *p, int repetition)
 {
   unsigned int linenum;
-  unsigned int last_line_to_save = p->lines_required * (repetition + 1);
+  uintmax_t last_line_to_save = p->lines_required * (repetition + 1);
   struct cstring *line;
 
   create_output_file ();
@@ -1057,7 +996,7 @@ close_output_file (void)
 {
   if (output_stream)
     {
-      if (fclose (output_stream) == EOF)
+      if (ferror (output_stream) || fclose (output_stream) == EOF)
 	{
 	  error (0, errno, _("write error for `%s'"), output_filename);
 	  output_stream = NULL;
@@ -1070,8 +1009,12 @@ close_output_file (void)
 	  files_created--;
 	}
       else
-        if (!suppress_count)
-	  fprintf (stdout, "%d\n", bytes_written);
+	{
+	  /* FIXME: if we write to stdout here, we have to close stdout
+	     and check for errors.  */
+	  if (!suppress_count)
+	    fprintf (stdout, "%d\n", bytes_written);
+	}
       output_stream = NULL;
     }
 }
@@ -1146,7 +1089,7 @@ check_for_offset (struct control *p, const char *str, const char *num)
 static void
 parse_repeat_count (int argnum, struct control *p, char *str)
 {
-  unsigned long val;
+  uintmax_t val;
   char *end;
 
   end = str + strlen (str) - 1;
@@ -1158,14 +1101,13 @@ parse_repeat_count (int argnum, struct control *p, char *str)
     p->repeat_forever = 1;
   else
     {
-      if (xstrtoul (str + 1, NULL, 10, &val, "") != LONGINT_OK
-	  || val > UINT_MAX)
+      if (xstrtoumax (str + 1, NULL, 10, &val, "") != LONGINT_OK)
 	{
 	  error (EXIT_FAILURE, 0,
 		 _("%s}: integer required between `{' and `}'"),
 		 global_argv[argnum]);
 	}
-      p->repeat = (unsigned int) val;
+      p->repeat = val;
     }
 
   *end = '}';
@@ -1202,9 +1144,6 @@ extract_regexp (int argnum, boolean ignore, char *str)
   p->re_compiled.buffer = (unsigned char *) xmalloc (p->re_compiled.allocated);
   p->re_compiled.fastmap = xmalloc (256);
   p->re_compiled.translate = 0;
-#if !WITH_REGEX
-  p->re_compiled.syntax_parens = 0;
-#endif
   err = re_compile_pattern (p->regexpr, len, &p->re_compiled);
   if (err)
     {
@@ -1226,8 +1165,8 @@ parse_patterns (int argc, int start, char **argv)
 {
   int i;			/* Index into ARGV. */
   struct control *p;		/* New control record created. */
-  unsigned long val;
-  static unsigned long last_val = 0;
+  uintmax_t val;
+  static uintmax_t last_val = 0;
 
   for (i = start; i < argc; i++)
     {
@@ -1240,25 +1179,28 @@ parse_patterns (int argc, int start, char **argv)
 	  p = new_control_record ();
 	  p->argnum = i;
 
-	  if (xstrtoul (argv[i], NULL, 10, &val, "") != LONGINT_OK
-	      || val > INT_MAX)
+	  if (xstrtoumax (argv[i], NULL, 10, &val, "") != LONGINT_OK)
 	    error (EXIT_FAILURE, 0, _("%s: invalid pattern"), argv[i]);
 	  if (val == 0)
 	    error (EXIT_FAILURE, 0,
 		   _("%s: line number must be greater than zero"),
 		   argv[i]);
 	  if (val < last_val)
-	    error (EXIT_FAILURE, 0,
-	     _("line number `%s' is smaller than preceding line number, %lu"),
-		   argv[i], last_val);
+	    {
+	      char buf[LONGEST_HUMAN_READABLE + 1];
+	      error (EXIT_FAILURE, 0,
+	       _("line number `%s' is smaller than preceding line number, %s"),
+		     argv[i], human_readable (last_val, buf, 1, 1));
+	    }
 
 	  if (val == last_val)
 	    error (0, 0,
 	   _("warning: line number `%s' is the same as preceding line number"),
 		   argv[i]);
+
 	  last_val = val;
 
-	  p->lines_required = (int) val;
+	  p->lines_required = val;
 	}
 
       if (i + 1 < argc && *argv[i + 1] == '{')
@@ -1429,7 +1371,7 @@ main (int argc, char **argv)
 {
   int optc;
   unsigned long val;
-#ifdef SA_INTERRUPT
+#ifdef SA_NOCLDSTOP
   struct sigaction oldact, newact;
 #endif
 
@@ -1438,6 +1380,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   global_argv = argv;
   controls = NULL;
   control_used = 0;
@@ -1445,7 +1389,10 @@ main (int argc, char **argv)
   remove_files = TRUE;
   prefix = DEFAULT_PREFIX;
 
-#ifdef SA_INTERRUPT
+  /* Change the way xmalloc and xrealloc fail.  */
+  xalloc_fail_func = cleanup;
+
+#ifdef SA_NOCLDSTOP
   newact.sa_handler = interrupt_handler;
   sigemptyset (&newact.sa_mask);
   newact.sa_flags = 0;
@@ -1465,7 +1412,7 @@ main (int argc, char **argv)
   sigaction (SIGTERM, NULL, &oldact);
   if (oldact.sa_handler != SIG_IGN)
     sigaction (SIGTERM, &newact, NULL);
-#else /* not SA_INTERRUPT */
+#else
   if (signal (SIGHUP, SIG_IGN) != SIG_IGN)
     signal (SIGHUP, interrupt_handler);
   if (signal (SIGINT, SIG_IGN) != SIG_IGN)
@@ -1474,10 +1421,9 @@ main (int argc, char **argv)
     signal (SIGQUIT, interrupt_handler);
   if (signal (SIGTERM, SIG_IGN) != SIG_IGN)
     signal (SIGTERM, interrupt_handler);
-#endif /* not SA_INTERRUPT */
+#endif
 
-  while ((optc = getopt_long (argc, argv, "f:b:kn:sqz", longopts, (int *) 0))
-	 != EOF)
+  while ((optc = getopt_long (argc, argv, "f:b:kn:sqz", longopts, NULL)) != -1)
     switch (optc)
       {
       case 0:
@@ -1511,19 +1457,14 @@ main (int argc, char **argv)
 	elide_empty_files = TRUE;
 	break;
 
+      case_GETOPT_HELP_CHAR;
+
+      case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
+
       default:
 	usage (1);
       }
 
-  if (show_version)
-    {
-      printf ("csplit (%s) %s\n", GNU_PACKAGE, VERSION);
-      exit (EXIT_SUCCESS);
-    }
-
-  if (show_help)
-    usage (0);
-
   if (argc - optind < 2)
     {
       error (0, 0, _("too few arguments"));
@@ -1550,7 +1491,7 @@ main (int argc, char **argv)
   exit (EXIT_SUCCESS);
 }
 
-static void
+void
 usage (int status)
 {
   if (status != 0)
@@ -1585,7 +1526,7 @@ Read standard input if FILE is -.  Each PATTERN may be:\n\
 \n\
 A line OFFSET is a required `+' or `-' followed by a positive integer.\n\
 "));
-      puts (_("\nReport bugs to textutils-bugs@gnu.ai.mit.edu"));
+      puts (_("\nReport bugs to <bug-textutils@gnu.org>."));
     }
   exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
