@@ -1,5 +1,5 @@
 /* join - join lines of two files on a common field
-   Copyright (C) 91, 1995-1999 Free Software Foundation, Inc.
+   Copyright (C) 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 #include <getopt.h>
 
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
 #include "hard-locale.h"
 #include "linebuffer.h"
@@ -708,11 +709,12 @@ void
 make_blank (struct line *blank, int count)
 {
   int i;
-  char *buffer;
+  unsigned char *buffer;
   struct field *fields;
   blank->nfields = count;
   blank->buf.size = blank->buf.length = count + 1;
-  blank->buf.buffer = buffer = xmalloc (blank->buf.size);
+  blank->buf.buffer = xmalloc (blank->buf.size);
+  buffer = (unsigned char *) blank->buf.buffer;
   blank->fields = fields =
     (struct field *) xmalloc (sizeof (struct field) * count);
   for (i = 0; i < count; i++)
@@ -736,6 +738,8 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
 #ifdef ENABLE_NLS
   hard_LC_COLLATE = hard_locale (LC_COLLATE);
 #endif
@@ -867,8 +871,6 @@ main (int argc, char **argv)
     error (EXIT_FAILURE, errno, "%s", names[1]);
   if ((fp1 == stdin || fp2 == stdin) && fclose (stdin) == EOF)
     error (EXIT_FAILURE, errno, "-");
-  if (ferror (stdout) || fclose (stdout) == EOF)
-    error (EXIT_FAILURE, errno, _("write error"));
 
   exit (EXIT_SUCCESS);
 }
