@@ -1,5 +1,5 @@
 /* od -- dump files in octal and other formats
-   Copyright (C) 92, 1995-1999 Free Software Foundation, Inc.
+   Copyright (C) 92, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
 #include <getopt.h>
 #include <sys/types.h>
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
 #include "xstrtol.h"
 
@@ -76,6 +77,8 @@ enum size_spec
     SHORT,
     INT,
     LONG,
+    LONG_LONG,
+    /* FIXME: add INTMAX support, too */
     FLOAT_SINGLE,
     FLOAT_DOUBLE,
     FLOAT_LONG_DOUBLE
@@ -228,7 +231,11 @@ static FILE *in_stream;
 /* If nonzero, at least one of the files we read was standard input.  */
 static int have_read_stdin;
 
-#define LONGEST_INTEGRAL_TYPE long int
+#ifdef HAVE_UNSIGNED_LONG_LONG
+# define LONGEST_INTEGRAL_TYPE unsigned long long
+#else
+# define LONGEST_INTEGRAL_TYPE long int
+#endif
 
 #define MAX_INTEGRAL_TYPE_SIZE sizeof(LONGEST_INTEGRAL_TYPE)
 static enum size_spec integral_type_size[MAX_INTEGRAL_TYPE_SIZE + 1];
@@ -440,6 +447,21 @@ print_long (long unsigned int n_bytes, const char *block,
     }
 }
 
+#ifdef HAVE_UNSIGNED_LONG_LONG
+static void
+print_long_long (long unsigned int n_bytes, const char *block,
+		 const char *fmt_string)
+{
+  int i;
+  for (i = n_bytes / sizeof (unsigned long long); i > 0; i--)
+    {
+      unsigned long long tmp = *(const unsigned long long *) block;
+      printf (fmt_string, tmp);
+      block += sizeof (unsigned long long);
+    }
+}
+#endif
+
 static void
 print_float (long unsigned int n_bytes, const char *block,
 	     const char *fmt_string)
@@ -703,7 +725,9 @@ this system doesn't provide a %lu-byte integral type"), s_orig, size);
 	  fmt = SIGNED_DECIMAL;
 	  sprintf (fmt_string, " %%%u%sd",
 		   (field_width = bytes_to_signed_dec_digits[size]),
-		   (size_spec == LONG ? "l" : ""));
+		   (size_spec == LONG ? "l"
+		    : (size_spec == LONG_LONG ? "ll"
+		       : "")));
 	  break;
 
 	case 'o':
@@ -755,6 +779,10 @@ this system doesn't provide a %lu-byte integral type"), s_orig, size);
 	  print_function = print_long;
 	  break;
 
+	case LONG_LONG:
+	  print_function = print_long_long;
+	  break;
+
 	default:
 	  abort ();
 	}
@@ -1362,15 +1390,11 @@ dump (void)
 {
   char *block[2];
   off_t current_offset;
-  off_t end_offset;
+  off_t end_offset IF_LINT (= 0);
   int idx;
   int err;
   size_t n_bytes_read;
 
-#ifdef lint  /* Suppress `used before initialized' warning.  */
-  end_offset = 0;
-#endif
-
   block[0] = (char *) alloca (bytes_per_block);
   block[1] = (char *) alloca (bytes_per_block);
 
@@ -1581,17 +1605,15 @@ main (int argc, char **argv)
 
   /* The old-style `pseudo starting address' to be printed in parentheses
      after any true address.  */
-  long int pseudo_start;
-
-#ifdef lint  /* Suppress `used before initialized' warning.  */
-  pseudo_start = 0;
-#endif
+  long int pseudo_start IF_LINT (= 0);
 
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   err = 0;
 
   for (i = 0; i <= MAX_INTEGRAL_TYPE_SIZE; i++)
@@ -1601,6 +1623,9 @@ main (int argc, char **argv)
   integral_type_size[sizeof (short int)] = SHORT;
   integral_type_size[sizeof (int)] = INT;
   integral_type_size[sizeof (long int)] = LONG;
+#ifdef HAVE_UNSIGNED_LONG_LONG
+  integral_type_size[sizeof (long long)] = LONG_LONG;
+#endif
 
   for (i = 0; i <= MAX_FP_TYPE_SIZE; i++)
     fp_type_size[i] = NO_SIZE;
@@ -1936,8 +1961,5 @@ cleanup:;
   if (have_read_stdin && fclose (stdin) == EOF)
     error (EXIT_FAILURE, errno, _("standard input"));
 
-  if (fclose (stdout) == EOF)
-    error (EXIT_FAILURE, errno, _("write error"));
-
   exit (err == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 }
