@@ -1,5 +1,5 @@
 /* sort - sort lines of text (with all kinds of options).
-   Copyright (C) 88, 1991-1999 Free Software Foundation, Inc.
+   Copyright (C) 88, 1991-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,6 +28,7 @@
 #include <stdio.h>
 #include <assert.h>
 #include "system.h"
+#include "closeout.h"
 #include "long-options.h"
 #include "error.h"
 #include "hard-locale.h"
@@ -53,8 +54,6 @@
 double strtod ();
 #endif
 
-char *xstrdup ();
-
 /* Undefine, to avoid warning about redefinition on some systems.  */
 #undef min
 #define min(a, b) ((a) < (b) ? (a) : (b))
@@ -74,6 +73,7 @@ char *xstrdup ();
    not properly sorted.  Any other irregular exit must exit with a
    status code greater than 1.  */
 #define SORT_FAILURE 2
+#define SORT_OUT_OF_ORDER 1
 
 #define C_DECIMAL_POINT '.'
 #define NEGATION_SIGN   '-'
@@ -86,7 +86,6 @@ static int th_sep; /* if CHAR_MAX + 1, then there is no thousands separator */
 
 /* Nonzero if the corresponding locales are hard.  */
 static int hard_LC_COLLATE;
-static int hard_LC_CTYPE;
 # if HAVE_NL_LANGINFO
 static int hard_LC_TIME;
 # endif
@@ -130,7 +129,9 @@ struct buffer
   char *buf;			/* Dynamically allocated buffer. */
   int used;			/* Number of bytes used. */
   int alloc;			/* Number of bytes allocated. */
-  int left;			/* Number of bytes left after line parsing. */
+  int left;			/* Number of bytes left from previous reads. */
+  int newline_free;		/* Number of left bytes that are known
+                                   to be newline-free.  */
 };
 
 struct keyfield
@@ -162,6 +163,13 @@ struct month
 /* The name this program was run with. */
 char *program_name;
 
+/* FIXME: None of these tables work with multibyte character sets.
+   Also, there are many other bugs when handling multibyte characters,
+   or even unibyte encodings where line boundaries are not in the
+   initial shift state.  One way to fix this is to rewrite `sort' to
+   use wide characters internally, but doing this with good
+   performance is a bit tricky.  */
+
 /* Table of white space. */
 static int blanks[UCHAR_LIM];
 
@@ -171,8 +179,7 @@ static int nonprinting[UCHAR_LIM];
 /* Table of non-dictionary characters (not letters, digits, or blanks). */
 static int nondictionary[UCHAR_LIM];
 
-/* Translation table folding lower case to upper.
-   FIXME: This doesn't work with multibyte character sets.  */
+/* Translation table folding lower case to upper.  */
 static char fold_toupper[UCHAR_LIM];
 
 #define MONTHS_PER_YEAR 12
@@ -204,22 +211,32 @@ static MONTHTAB_CONST struct month monthtab[] =
 /* During the merge phase, the number of files to merge at once. */
 #define NMERGE 16
 
-/* Initial buffer size for in core sorting.  Will not grow unless a
-   line longer than this is seen. */
-static int sortalloc = 512 * 1024 + 1;
+/* Initial buffer size for in-core sorting.  The buffer will grow only
+   if a line longer than this is seen. */
+#define SORTALLOC (8 * 1024 * 1024)
+static int const sortalloc = SORTALLOC;
 
 /* Initial buffer size for in core merge buffers.  Bear in mind that
    up to NMERGE * mergealloc bytes may be allocated for merge buffers. */
-static int mergealloc = 16 * 1024 + 1;
+static int const mergealloc = SORTALLOC / NMERGE / 2;
 
 /* Guess of average line length. */
-static int linelength = 30;
+static int const linelength = 30;
 
 /* Maximum number of elements for the array(s) of struct line's, in bytes.  */
-#define LINEALLOC (256 * 1024)
+#define LINEALLOC (SORTALLOC / 2)
 
-/* Directory in which any temporary files are to be created. */
-static char *temp_dir;
+/* Array of directory names in which any temporary files are to be created. */
+static char const **temp_dirs;
+
+/* Number of temporary directory names used.  */
+static size_t temp_dir_count;
+
+/* Number of allocated slots in temp_dirs.  */
+static size_t temp_dir_alloc;
+
+/* Our process ID.  */
+static pid_t process_id;
 
 /* Flag to reverse the order of all comparisons. */
 static int reverse;
@@ -241,8 +258,8 @@ static int unique;
 /* Nonzero if any of the input files are the standard input. */
 static int have_read_stdin;
 
-/* Lists of key field comparisons to be tried. */
-static struct keyfield keyhead;
+/* List of key field comparisons to be tried.  */
+static struct keyfield *keylist;
 
 void
 usage (int status)
@@ -264,10 +281,10 @@ Write sorted concatenation of all FILE(s) to standard output.\n\
                      starting with zero (contrast with the -k option)\n\
   -b               ignore leading blanks in sort fields or keys\n\
   -c               check if given files already sorted, do not sort\n\
-  -d               consider only [a-zA-Z0-9 ] characters in keys\n\
+  -d               consider only blanks and alphanumeric characters in keys\n\
   -f               fold lower case to upper case characters in keys\n\
   -g               compare according to general numerical value, imply -b\n\
-  -i               consider only [\\040-\\0176] characters in keys\n\
+  -i               consider only printable characters in keys\n\
   -k POS1[,POS2]   start a key at POS1, end it *at* POS2\n\
 		     field numbers and character offsets are numbered\n\
                      starting with one (contrast with zero-based +POS form)\n\
@@ -279,6 +296,7 @@ Write sorted concatenation of all FILE(s) to standard output.\n\
   -s               stabilize sort by disabling last resort comparison\n\
   -t SEP           use SEParator instead of non- to whitespace transition\n\
   -T DIRECTORY     use DIRECTORY for temporary files, not $TMPDIR or %s\n\
+		     multiple -T options specify multiple directories\n\
   -u               with -c, check for strict ordering;\n\
                    with -m, only output the first of an equal sequence\n\
   -z               end lines with 0 byte, not newline, for find -print0\n\
@@ -293,6 +311,10 @@ in the field, both counted from one with -k, from zero with the obsolescent\n\
 form.  OPTS is made up of one or more of Mbdfinr; this effectively disables\n\
 global -Mbdfinr settings for that key.  If no key is given, use the entire\n\
 line as the key.  With no FILE, or when FILE is -, read standard input.\n\
+\n\
+*** WARNING ***\n\
+The locale specified by the environment affects sort order.\n\
+Set LC_ALL=C to get the traditional sort order that uses native byte values.\n\
 ")
 	      );
       puts (_("\nReport bugs to <bug-textutils@gnu.org>."));
@@ -405,37 +427,46 @@ write_bytes (const char *buf, size_t n_bytes, FILE *fp, const char *output_file)
     }
 }
 
+/* Append DIR to the array of temporary directory names.  */
+static void
+add_temp_dir (char const *dir)
+{
+  if (temp_dir_count == temp_dir_alloc)
+    {
+      temp_dir_alloc = temp_dir_alloc ? temp_dir_alloc * 2 : 16;
+      temp_dirs = xrealloc (temp_dirs, sizeof (temp_dirs) * temp_dir_alloc);
+    }
+
+  temp_dirs[temp_dir_count++] = dir;
+}
+
 /* Return a name for a temporary file. */
 
 static char *
 tempname (void)
 {
-  static unsigned int seq;
-  int len = strlen (temp_dir);
-  char *name = xmalloc (len + 1 + sizeof ("sort") - 1 + 5 + 5 + 1);
+  static unsigned long sequence_number;
+
+  unsigned long seq = sequence_number++;
+  unsigned long pid = process_id;
+  char const *temp_dir = temp_dirs[seq % temp_dir_count];
+  size_t len = strlen (temp_dir);
+  char const *slash = "/" + (len == 0 || temp_dir[len - 1] == '/');
+  char *name = xmalloc (len + 1 + sizeof "sort" - 1
+			+ sizeof pid * CHAR_BIT / 3 + 1
+			+ sizeof seq * CHAR_BIT / 3 + 1);
   int long_file_names = NAME_MAX_IN_DIR (temp_dir) > 12;
   struct tempnode *node;
 
-  /* If long filenames aren't supported, we cannot use filenames
-     longer than 8+3 and still assume they are unique.  */
   if (long_file_names)
-    sprintf (name,
-	     "%s%ssort%5.5d%5.5d",
-	     temp_dir,
-	     (len && temp_dir[len - 1] != '/') ? "/" : "",
-	     (unsigned int) getpid () & 0xffff, seq);
+    sprintf (name, "%s%ssort%lu.%.5lu", temp_dir, slash, pid, seq);
   else
-    sprintf (name, "%s%ss%5.5d%2.2d.%3.3d",
-	     temp_dir,
-	     (len && temp_dir[len - 1] != '/') ? "/" : "",
-	     (unsigned int) getpid () & 0xffff, seq / 1000, seq % 1000);
-
-  ++seq;
-
-  /* Make sure that SEQ's value fits in 5 digits if temp_dir is on
-     an 8.3 filesystem.  */
-  if (!long_file_names && seq >= 100000)
-    seq = 0;
+    {
+      /* Make sure the file name is safe for an 8.3 filesystem.  */
+      sprintf (name, "%s%ss%.5d%.2d.%.3d", temp_dir, slash,
+	       (int) (pid % 100000), (int) (seq / 1000 % 100),
+	       (int) (seq % 1000));
+    }
 
   node = (struct tempnode *) xmalloc (sizeof (struct tempnode));
   node->name = name;
@@ -530,32 +561,32 @@ initbuf (struct buffer *buf, int alloc)
 {
   buf->alloc = alloc;
   buf->buf = xmalloc (buf->alloc);
-  buf->used = buf->left = 0;
+  buf->used = buf->left = buf->newline_free = 0;
 }
 
 /* Fill BUF reading from FP, moving buf->left bytes from the end
    of buf->buf to the beginning first.  If EOF is reached and the
-   file wasn't terminated by a newline, supply one.  Always leave
-   at least one unused byte at the end.  Return a count of bytes
-   buffered. */
+   file wasn't terminated by a newline, supply one.  Return a count
+   of bytes buffered. */
 
 static int
 fillbuf (struct buffer *buf, FILE *fp)
 {
-  int cc;
-
-  memmove (buf->buf, buf->buf + buf->used - buf->left, buf->left);
-  buf->used = buf->left;
-
-  while (!feof (fp) && (buf->used == 0
-			|| !memchr (buf->buf, eolchar, buf->used)))
+  if (buf->used != buf->left)
     {
-      if (buf->used == buf->alloc - 1)
+      memmove (buf->buf, buf->buf + buf->used - buf->left, buf->left);
+      buf->used = buf->left;
+    }
+
+  while (!feof (fp))
+    {
+      int cc;
+      if (buf->used == buf->alloc)
 	{
-	  buf->alloc = buf->alloc * 2 - 1;
+	  buf->alloc *= 2;
 	  buf->buf = xrealloc (buf->buf, buf->alloc);
 	}
-      cc = fread (buf->buf + buf->used, 1, buf->alloc - 1 - buf->used, fp);
+      cc = fread (buf->buf + buf->used, 1, buf->alloc - buf->used, fp);
       if (ferror (fp))
 	{
 	  error (0, errno, _("read error"));
@@ -563,13 +594,20 @@ fillbuf (struct buffer *buf, FILE *fp)
 	  exit (SORT_FAILURE);
 	}
       buf->used += cc;
+      if (memchr (buf->buf + buf->used - cc, eolchar, cc))
+	break;
     }
 
   if (feof (fp) && buf->used && buf->buf[buf->used - 1] != eolchar)
     {
-      if (buf->used == buf->alloc - 1)
+      if (buf->used == buf->alloc)
 	{
-	  buf->alloc = buf->alloc * 2 - 1;
+	  /* If the buffer isn't parsed yet, don't increase the buffer
+	     size to append a newline, as this won't be needed unless
+	     the buffer turns out to be newline-free.  */
+	  if (buf->newline_free != buf->used)
+	    return buf->used;
+	  buf->alloc *= 2;
 	  buf->buf = xrealloc (buf->buf, buf->alloc);
 	}
       buf->buf[buf->used++] = eolchar;
@@ -597,7 +635,7 @@ initlines (struct lines *lines, int alloc, int limit)
 static char *
 begfield (const struct line *line, const struct keyfield *key)
 {
-  register char *ptr = line->text, *lim = ptr + line->length;
+  register char *ptr = line->text, *lim = ptr + line->length - 1;
   register int sword = key->sword, schar = key->schar;
 
   if (tab)
@@ -635,7 +673,7 @@ begfield (const struct line *line, const struct keyfield *key)
 static char *
 limfield (const struct line *line, const struct keyfield *key)
 {
-  register char *ptr = line->text, *lim = ptr + line->length;
+  register char *ptr = line->text, *lim = ptr + line->length - 1;
   register int eword = key->eword, echar = key->echar;
 
   /* Note: from the POSIX spec:
@@ -746,13 +784,15 @@ static void
 findlines (struct buffer *buf, struct lines *lines)
 {
   register char *beg = buf->buf, *lim = buf->buf + buf->used, *ptr;
-  struct keyfield *key = keyhead.next;
+  struct keyfield *key = keylist;
 
   lines->used = 0;
 
   while (beg < lim && (ptr = memchr (beg, eolchar, lim - beg))
 	 && lines->used < lines->limit)
     {
+      struct line *line;
+
       if (lines->used == lines->alloc)
 	{
 	  lines->alloc *= 2;
@@ -761,46 +801,33 @@ findlines (struct buffer *buf, struct lines *lines)
 		      lines->alloc * sizeof (struct line));
 	}
 
-      ptr++;
-      lines->lines[lines->used].text = beg;
-      lines->lines[lines->used].length = ptr - beg;
+      line = &lines->lines[lines->used];
+      line->text = beg;
+      line->length = ptr + 1 - beg;
 
-      /* Precompute the position of the first key for efficiency. */
       if (key)
 	{
-	  if (key->eword >= 0)
-	    lines->lines[lines->used].keylim =
-	      limfield (&lines->lines[lines->used], key);
-	  else
-	    lines->lines[lines->used].keylim = ptr;
+	  /* Precompute the position of the first key for efficiency. */
+	  line->keylim = 0 <= key->eword ? limfield (line, key) : ptr;
 
 	  if (key->sword >= 0)
-	    lines->lines[lines->used].keybeg =
-	      begfield (&lines->lines[lines->used], key);
+	    line->keybeg = begfield (line, key);
 	  else
 	    {
 	      if (key->skipsblanks)
 		while (blanks[UCHAR (*beg)])
 		  ++beg;
-	      lines->lines[lines->used].keybeg = beg;
+	      line->keybeg = beg;
 	    }
 	  if (key->skipeblanks)
-	    {
-	      trim_trailing_blanks (lines->lines[lines->used].keybeg,
-				    &lines->lines[lines->used].keylim);
-	    }
-	}
-      else
-	{
-	  lines->lines[lines->used].keybeg = 0;
-	  lines->lines[lines->used].keylim = 0;
+	    trim_trailing_blanks (line->keybeg, &line->keylim);
 	}
 
       ++lines->used;
-      beg = ptr;
+      beg = ptr + 1;
     }
 
-  buf->left = lim - beg;
+  buf->newline_free = buf->left = lim - beg;
 }
 
 /* Compare strings A and B containing decimal fractions < 1.  Each string
@@ -1087,48 +1114,21 @@ getmonth (const char *s, int len)
 static int
 keycompare (const struct line *a, const struct line *b)
 {
-  register char *texta, *textb, *lima, *limb;
-  register unsigned char *translate;
-  register int *ignore;
-  struct keyfield *key;
-  int diff = 0, iter = 0, lena, lenb;
+  struct keyfield *key = keylist;
 
-  for (key = keyhead.next; key; key = key->next, ++iter)
+  /* For the first iteration only, the key positions have been
+     precomputed for us. */
+  register char *texta = a->keybeg;
+  register char *textb = b->keybeg;
+  register char *lima = a->keylim;
+  register char *limb = b->keylim;
+
+  int diff, lena, lenb;
+
+  for (;;)
     {
-      int comparable_lengths = 1;
-
-      ignore = key->ignore;
-      translate = (unsigned char *) key->translate;
-
-      /* Find the beginning and limit of each field. */
-      if (iter || a->keybeg == NULL || b->keybeg == NULL)
-	{
-	  if (key->eword >= 0)
-	    lima = limfield (a, key), limb = limfield (b, key);
-	  else
-	    lima = a->text + a->length, limb = b->text + b->length;
-
-	  if (key->sword >= 0)
-	    texta = begfield (a, key), textb = begfield (b, key);
-	  else
-	    {
-	      texta = a->text, textb = b->text;
-	      if (key->skipsblanks)
-		{
-		  while (texta < lima && blanks[UCHAR (*texta)])
-		    ++texta;
-		  while (textb < limb && blanks[UCHAR (*textb)])
-		    ++textb;
-		}
-	    }
-	}
-      else
-	{
-	  /* For the first iteration only, the key positions have
-	     been precomputed for us. */
-	  texta = a->keybeg, lima = a->keylim;
-	  textb = b->keybeg, limb = b->keylim;
-	}
+      register unsigned char *translate = (unsigned char *) key->translate;
+      register int *ignore = key->ignore;
 
       /* Find the lengths. */
       lena = lima - texta, lenb = limb - textb;
@@ -1150,44 +1150,24 @@ keycompare (const struct line *a, const struct line *b)
       /* Actually compare the fields. */
       if (key->numeric | key->general_numeric)
 	{
-	  char savea, saveb;
+	  char savea = *lima, saveb = *limb;
 
-	  /* If the fields are adjacent, adjust the end of the earlier
-	     field back by 1 byte, since we temporarily modify the
-	     byte after the field during comparison.  This can't
-	     change a numeric comparison, since the byte is a newline.
-	     If the earlier field is empty, adjust its start as well.  */
-	  if (lima == textb)
-	    texta -= texta == lima--;
-	  if (limb == texta)
-	    textb -= textb == limb--;
-
-	  savea = *lima, saveb = *limb;
 	  *lima = *limb = '\0';
 	  diff = ((key->numeric ? numcompare : general_numcompare)
 		  (texta, textb));
 	  *lima = savea, *limb = saveb;
-	  if (diff)
-	    return key->reverse ? -diff : diff;
-	  continue;
 	}
       else if (key->month)
-	{
-	  diff = getmonth (texta, lena) - getmonth (textb, lenb);
-	  if (diff)
-	    return key->reverse ? -diff : diff;
-	  continue;
-	}
+	diff = getmonth (texta, lena) - getmonth (textb, lenb);
 #ifdef ENABLE_NLS
-
       /* Sorting like this may become slow, so in a simple locale the user
          can select a faster sort that is similar to ascii sort  */
-      else if (hard_LC_COLLATE | hard_LC_CTYPE)
+      else if (hard_LC_COLLATE)
 	{
 	  if (ignore || translate)
 	    {
-	      char *copy_a = (char *) alloca (lena + 1);
-	      char *copy_b = (char *) alloca (lenb + 1);
+	      char *copy_a = (char *) alloca (lena + 1 + lenb + 1);
+	      char *copy_b = copy_a + lena + 1;
 	      int new_len_a, new_len_b, i;
 
 	      /* Ignore and/or translate chars before comparing.  */
@@ -1213,103 +1193,103 @@ keycompare (const struct line *a, const struct line *b)
 
 	      diff = memcoll (copy_a, new_len_a, copy_b, new_len_b);
 	    }
+	  else if (lena == 0)
+	    diff = -lenb;
+	  else if (lenb == 0)
+	    goto greater;
 	  else
-	    {
-	      diff = memcoll (texta, lena, textb, lenb);
-	    }
-
-	  if (diff)
-	    return key->reverse ? -diff : diff;
-
-	  continue;
+	    diff = memcoll (texta, lena, textb, lenb);
 	}
 #endif
-      else if (ignore && translate)
-
+      else if (ignore)
+	{
 #define CMP_WITH_IGNORE(A, B)						\
   do									\
     {									\
-	  while (texta < lima && textb < limb)				\
+	  for (;;)							\
 	    {								\
 	      while (texta < lima && ignore[UCHAR (*texta)])		\
 		++texta;						\
 	      while (textb < limb && ignore[UCHAR (*textb)])		\
 		++textb;						\
-	      if (texta < lima && textb < limb)				\
-		{							\
-		  if ((A) != (B))					\
-		    {							\
-		      diff = UCHAR (A) - UCHAR (B);	      		\
-		      break;						\
-		    }							\
-		  ++texta;						\
-		  ++textb;						\
-		}							\
-									\
-	      if (texta == lima && textb < limb && !ignore[UCHAR (*textb)]) \
-		diff = -1;						\
-	      else if (texta < lima && textb == limb			\
-		       && !ignore[UCHAR (*texta)])			\
-		diff = 1;						\
+	      if (! (texta < lima && textb < limb))			\
+		break;							\
+	      diff = UCHAR (A) - UCHAR (B);				\
+	      if (diff)							\
+		goto not_equal;						\
+	      ++texta;							\
+	      ++textb;							\
 	    }								\
 									\
-	  if (diff == 0)						\
-	    {								\
-	      while (texta < lima && ignore[UCHAR (*texta)])		\
-		++texta;						\
-	      while (textb < limb && ignore[UCHAR (*textb)])		\
-		++textb;						\
-									\
-	      if (texta == lima && textb < limb)			\
-		diff = -1;						\
-	      else if (texta < lima && textb == limb)			\
-		diff = 1;						\
-	    }								\
-	  /* Relative lengths are meaningless if characters were ignored.  \
-	     Handling this case here avoids what might be an invalid length  \
-	     comparison below.  */					\
-	  if (diff == 0 && texta == lima && textb == limb)		\
-	    comparable_lengths = 0;					\
+	  diff = (lima - texta) - (limb - textb);			\
     }									\
   while (0)
 
-	CMP_WITH_IGNORE (translate[UCHAR (*texta)], translate[UCHAR (*textb)]);
-      else if (ignore)
-	CMP_WITH_IGNORE (UCHAR (*texta), UCHAR (*textb));
-      else if (translate)
-	while (texta < lima && textb < limb)
-	  {
-	    if (translate[UCHAR (*texta++)] != translate[UCHAR (*textb++)])
-	      {
-		diff = (UCHAR (translate[UCHAR (*--texta)])
-			- UCHAR (translate[UCHAR (*--textb)]));
-		break;
-	      }
-	  }
+	  if (translate)
+	    CMP_WITH_IGNORE (translate[UCHAR (*texta)],
+			     translate[UCHAR (*textb)]);
+	  else
+	    CMP_WITH_IGNORE (UCHAR (*texta), UCHAR (*textb));
+	}
+      else if (lena == 0)
+	diff = -lenb;
+      else if (lenb == 0)
+	goto greater;
       else
 	{
-#ifdef ENABLE_NLS
-	  if (hard_LC_COLLATE)
+	  if (translate)
 	    {
-	      /* Ignore any length difference if the localized comparison
-		 says the strings are equal.  */
-	      comparable_lengths = 0;
-	      diff = memcoll (texta, lena, textb, lenb);
+	      while (texta < lima && textb < limb)
+		{
+		  diff = (UCHAR (translate[UCHAR (*texta++)])
+			  - UCHAR (translate[UCHAR (*textb++)]));
+		  if (diff)
+		    goto not_equal;
+		}
 	    }
 	  else
-#endif
 	    {
 	      diff = memcmp (texta, textb, min (lena, lenb));
+	      if (diff)
+		goto not_equal;
 	    }
+	  diff = lena - lenb;
 	}
 
       if (diff)
-	return key->reverse ? -diff : diff;
-      if (comparable_lengths && (diff = lena - lenb) != 0)
-	return key->reverse ? -diff : diff;
+	goto not_equal;
+
+      key = key->next;
+      if (! key)
+	break;
+
+      /* Find the beginning and limit of the next field.  */
+      if (key->eword >= 0)
+	lima = limfield (a, key), limb = limfield (b, key);
+      else
+	lima = a->text + a->length - 1, limb = b->text + b->length - 1;
+
+      if (key->sword >= 0)
+	texta = begfield (a, key), textb = begfield (b, key);
+      else
+	{
+	  texta = a->text, textb = b->text;
+	  if (key->skipsblanks)
+	    {
+	      while (texta < lima && blanks[UCHAR (*texta)])
+		++texta;
+	      while (textb < limb && blanks[UCHAR (*textb)])
+		++textb;
+	    }
+	}
     }
 
   return 0;
+
+ greater:
+  diff = 1;
+ not_equal:
+  return key->reverse ? -diff : diff;
 }
 
 /* Compare two lines A and B, returning negative, zero, or positive
@@ -1318,41 +1298,33 @@ keycompare (const struct line *a, const struct line *b)
 static int
 compare (register const struct line *a, register const struct line *b)
 {
-  int diff, tmpa, tmpb;
+  int diff, alen, blen;
 
   /* First try to compare on the specified keys (if any).
      The only two cases with no key at all are unadorned sort,
      and unadorned sort -r. */
-  if (keyhead.next)
+  if (keylist)
     {
       diff = keycompare (a, b);
+      alloca (0);
       if (diff != 0 || unique || stable)
-	{
-	  alloca (0);
-	  return diff;
-	}
+	return diff;
     }
 
   /* If the keys all compare equal (or no keys were specified)
-     fall through to the default byte-by-byte comparison. */
-  tmpa = a->length, tmpb = b->length;
+     fall through to the default comparison.  */
+  alen = a->length - 1, blen = b->length - 1;
 
+  if (alen == 0)
+    diff = - blen;
+  else if (blen == 0)
+    diff = alen;
 #ifdef ENABLE_NLS
-  if (hard_LC_COLLATE)
-    {
-      diff = memcoll (a->text, tmpa, b->text, tmpb);
-      alloca (0);
-      return reverse ? -diff : diff;
-    }
+  else if (hard_LC_COLLATE)
+    diff = memcoll (a->text, alen, b->text, blen);
 #endif
-
-  diff = UCHAR (a->text[0]) - UCHAR (b->text[0]);
-  if (diff == 0)
-    {
-      diff = memcmp (a->text, b->text, min (tmpa, tmpb));
-      if (diff == 0)
-	diff = tmpa - tmpb;
-    }
+  else if (! (diff = memcmp (a->text, b->text, min (alen, blen))))
+    diff = alen - blen;
 
   return reverse ? -diff : diff;
 }
@@ -1371,12 +1343,9 @@ checkfp (FILE *fp, const char *file_name)
   int cc;			/* Character count. */
   int alloc;
   int line_number = 1;
-  struct line *disorder_line;
+  struct line *disorder_line IF_LINT (= NULL);
   int disorder_line_number = 0;
-
-#ifdef lint  /* Suppress `used before initialized' warning.  */
-  disorder_line = NULL;
-#endif
+  struct keyfield *key = keylist;
 
   initbuf (&buf, mergealloc);
   initlines (&lines, mergealloc / linelength + 1,
@@ -1412,20 +1381,19 @@ checkfp (FILE *fp, const char *file_name)
 
       /* Save the last line of the buffer and refill the buffer. */
       prev_line = lines.lines + (lines.used - 1);
-      if (prev_line->length + 1 > alloc)
+      if (alloc < prev_line->length)
 	{
-	  do
-	    {
-	      alloc *= 2;
-	    }
-	  while (alloc < prev_line->length + 1);
+	  while ((alloc *= 2) < prev_line->length)
+	    continue;
 	  temp.text = xrealloc (temp.text, alloc);
 	}
-      assert (prev_line->length + 1 <= alloc);
       memcpy (temp.text, prev_line->text, prev_line->length);
       temp.length = prev_line->length;
-      temp.keybeg = temp.text + (prev_line->keybeg - prev_line->text);
-      temp.keylim = temp.text + (prev_line->keylim - prev_line->text);
+      if (key)
+	{
+	  temp.keybeg = temp.text + (prev_line->keybeg - prev_line->text);
+	  temp.keylim = temp.text + (prev_line->keylim - prev_line->text);
+	}
 
       cc = fillbuf (&buf, fp);
       if (cc == 0)
@@ -1468,23 +1436,22 @@ mergefps (FILE **fps, register int nfps, FILE *ofp, const char *output_file)
 {
   struct buffer buffer[NMERGE];	/* Input buffers for each file. */
   struct lines lines[NMERGE];	/* Line tables for each buffer. */
-  struct line saved;		/* Saved line for unique check. */
-  int savedflag = 0;		/* True if there is a saved line. */
-  int savealloc;		/* Size allocated for the saved line. */
+  struct line saved;		/* Saved line storage for unique check. */
+  struct line const *savedline IF_LINT (= NULL);
+				/* &saved if there is a saved line. */
+  int savealloc IF_LINT (= 0);	/* Size allocated for the saved line. */
   int cur[NMERGE];		/* Current line in each line table. */
   int ord[NMERGE];		/* Table representing a permutation of fps,
 				   such that lines[ord[0]].lines[cur[ord[0]]]
 				   is the smallest line and will be next
 				   output. */
   register int i, j, t;
-
-#ifdef lint  /* Suppress `used before initialized' warning.  */
-  savealloc = 0;
-#endif
+  struct keyfield *key = keylist;
 
   /* Allocate space for a saved line if necessary. */
   if (unique)
     {
+      savedline = NULL;
       savealloc = linelength;
       saved.text = xmalloc (savealloc);
     }
@@ -1525,44 +1492,39 @@ mergefps (FILE **fps, register int nfps, FILE *ofp, const char *output_file)
   /* Repeatedly output the smallest line until no input remains. */
   while (nfps)
     {
+      struct line const *smallest = &lines[ord[0]].lines[cur[ord[0]]];
+
       /* If uniquified output is turned on, output only the first of
 	 an identical series of lines. */
       if (unique)
 	{
-	  if (savedflag && compare (&saved, &lines[ord[0]].lines[cur[ord[0]]]))
+	  if (savedline && compare (savedline, smallest))
 	    {
+	      savedline = 0;
 	      write_bytes (saved.text, saved.length, ofp, output_file);
-	      savedflag = 0;
 	    }
-	  if (!savedflag)
+	  if (!savedline)
 	    {
-	      if (savealloc < lines[ord[0]].lines[cur[ord[0]]].length + 1)
+	      savedline = &saved;
+	      if (savealloc < smallest->length)
 		{
-		  while (savealloc < lines[ord[0]].lines[cur[ord[0]]].length + 1)
-		    savealloc *= 2;
+		  while ((savealloc *= 2) < smallest->length)
+		    continue;
 		  saved.text = xrealloc (saved.text, savealloc);
 		}
-	      saved.length = lines[ord[0]].lines[cur[ord[0]]].length;
-	      memcpy (saved.text, lines[ord[0]].lines[cur[ord[0]]].text,
-		      saved.length);
-	      if (lines[ord[0]].lines[cur[ord[0]]].keybeg != NULL)
+	      saved.length = smallest->length;
+	      memcpy (saved.text, smallest->text, saved.length);
+	      if (key)
 		{
-		  saved.keybeg = saved.text +
-		    (lines[ord[0]].lines[cur[ord[0]]].keybeg
-		     - lines[ord[0]].lines[cur[ord[0]]].text);
+		  saved.keybeg =
+		    saved.text + (smallest->keybeg - smallest->text);
+		  saved.keylim =
+		    saved.text + (smallest->keylim - smallest->text);
 		}
-	      if (lines[ord[0]].lines[cur[ord[0]]].keylim != NULL)
-		{
-		  saved.keylim = saved.text +
-		    (lines[ord[0]].lines[cur[ord[0]]].keylim
-		     - lines[ord[0]].lines[cur[ord[0]]].text);
-		}
-	      savedflag = 1;
 	    }
 	}
       else
-	write_bytes (lines[ord[0]].lines[cur[ord[0]]].text,
-		     lines[ord[0]].lines[cur[ord[0]]].length, ofp, output_file);
+	write_bytes (smallest->text, smallest->length, ofp, output_file);
 
       /* Check if we need to read more lines into core. */
       if (++cur[ord[0]] == lines[ord[0]].used)
@@ -1613,7 +1575,7 @@ mergefps (FILE **fps, register int nfps, FILE *ofp, const char *output_file)
       ord[i - 1] = t;
     }
 
-  if (unique && savedflag)
+  if (unique && savedline)
     {
       write_bytes (saved.text, saved.length, ofp, output_file);
       free (saved.text);
@@ -1752,11 +1714,20 @@ sort (char **files, int nfiles, FILE *ofp, const char *output_file)
       fp = xfopen (*files++, "r");
       while (fillbuf (&buf, fp))
 	{
-	  findlines (&buf, &lines);
-	  if (lines.used > ntmp)
+	  if (nfiles && buf.used != buf.alloc && feof (fp))
 	    {
-	      while (lines.used > ntmp)
-		ntmp *= 2;
+	      /* End of file, but there is more input and buffer room.
+		 Concatenate the next input file; this is faster in
+		 the usual case.  */
+	      buf.left = buf.used;
+	      break;
+	    }
+
+	  findlines (&buf, &lines);
+	  if (ntmp < lines.used)
+	    {
+	      while ((ntmp *= 2) < lines.used)
+		continue;
 	      tmp = (struct line *)
 		xrealloc ((char *) tmp, ntmp * sizeof (struct line));
 	    }
@@ -1797,16 +1768,16 @@ sort (char **files, int nfiles, FILE *ofp, const char *output_file)
     }
 }
 
-/* Insert key KEY at the end of the list (`keyhead'). */
+/* Insert key KEY at the end of the key list.  */
 
 static void
 insertkey (struct keyfield *key)
 {
-  struct keyfield *k = &keyhead;
+  struct keyfield **p;
 
-  while (k->next)
-    k = k->next;
-  k->next = key;
+  for (p = &keylist; *p; p = &(*p)->next)
+    continue;
+  *p = key;
   key->next = NULL;
 }
 
@@ -1821,18 +1792,18 @@ badfieldspec (const char *s)
 static void
 sighandler (int sig)
 {
-#ifdef SA_INTERRUPT
+#ifdef SA_NOCLDSTOP
   struct sigaction sigact;
 
   sigact.sa_handler = SIG_DFL;
   sigemptyset (&sigact.sa_mask);
   sigact.sa_flags = 0;
   sigaction (sig, &sigact, NULL);
-#else				/* !SA_INTERRUPT */
+#else
   signal (sig, SIG_DFL);
-#endif				/* SA_INTERRUPT */
+#endif
   cleanup ();
-  kill (getpid (), sig);
+  kill (process_id, sig);
 }
 
 /* Set the ordering options for KEY specified in S.
@@ -1899,19 +1870,22 @@ main (int argc, char **argv)
   int checkonly = 0, mergeonly = 0, nfiles = 0;
   char *minus = "-", *outfile = minus, **files, *tmp;
   FILE *ofp;
-#ifdef SA_INTERRUPT
+#ifdef SA_NOCLDSTOP
   struct sigaction oldact, newact;
-#endif				/* SA_INTERRUPT */
+#endif
 
   program_name = argv[0];
+  process_id = getpid ();
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  close_stdout_set_status (SORT_FAILURE);
+  atexit (close_stdout);
+
 #ifdef ENABLE_NLS
 
   hard_LC_COLLATE = hard_locale (LC_COLLATE);
-  hard_LC_CTYPE = hard_locale (LC_CTYPE);
 # if HAVE_NL_LANGINFO
   hard_LC_TIME = hard_locale (LC_TIME);
 # endif
@@ -1941,15 +1915,11 @@ main (int argc, char **argv)
   have_read_stdin = 0;
   inittables ();
 
-  temp_dir = getenv ("TMPDIR");
-  if (temp_dir == NULL)
-    temp_dir = DEFAULT_TMPDIR;
-
   /* Change the way xmalloc and xrealloc fail.  */
   xalloc_exit_failure = SORT_FAILURE;
   xalloc_fail_func = cleanup;
 
-#ifdef SA_INTERRUPT
+#ifdef SA_NOCLDSTOP
   newact.sa_handler = sighandler;
   sigemptyset (&newact.sa_mask);
   newact.sa_flags = 0;
@@ -1966,7 +1936,7 @@ main (int argc, char **argv)
   sigaction (SIGTERM, NULL, &oldact);
   if (oldact.sa_handler != SIG_IGN)
     sigaction (SIGTERM, &newact, NULL);
-#else				/* !SA_INTERRUPT */
+#else
   if (signal (SIGINT, SIG_IGN) != SIG_IGN)
     signal (SIGINT, sighandler);
   if (signal (SIGHUP, SIG_IGN) != SIG_IGN)
@@ -1975,7 +1945,7 @@ main (int argc, char **argv)
     signal (SIGPIPE, sighandler);
   if (signal (SIGTERM, SIG_IGN) != SIG_IGN)
     signal (SIGTERM, sighandler);
-#endif				/* !SA_INTERRUPT */
+#endif
 
   gkey.sword = gkey.eword = -1;
   gkey.ignore = NULL;
@@ -2178,6 +2148,7 @@ but lacks following character offset"));
 			else
 			  outfile = argv[++i];
 		      }
+		    close_stdout_set_file_name (outfile);
 		    goto outer;
 		  case 's':
 		    stable = 1;
@@ -2196,11 +2167,11 @@ but lacks following character offset"));
 		    break;
 		  case 'T':
 		    if (s[1])
-		      temp_dir = ++s;
+		      add_temp_dir (++s);
 		    else
 		      {
 			if (i < argc - 1)
-			  temp_dir = argv[++i];
+			  add_temp_dir (argv[++i]);
 			else
 			  error (SORT_FAILURE, 0,
 				 _("option `-T' requires an argument"));
@@ -2237,7 +2208,7 @@ but lacks following character offset"));
     insertkey (key);
 
   /* Inheritance of global options to individual keys. */
-  for (key = keyhead.next; key; key = key->next)
+  for (key = keylist; key; key = key->next)
     if (!key->ignore && !key->translate && !key->skipsblanks && !key->reverse
 	&& !key->skipeblanks && !key->month && !key->numeric
         && !key->general_numeric)
@@ -2252,12 +2223,18 @@ but lacks following character offset"));
 	key->reverse = gkey.reverse;
       }
 
-  if (!keyhead.next && (gkey.ignore || gkey.translate || gkey.skipsblanks
-			|| gkey.skipeblanks || gkey.month || gkey.numeric
-                        || gkey.general_numeric))
+  if (!keylist && (gkey.ignore || gkey.translate || gkey.skipsblanks
+		   || gkey.skipeblanks || gkey.month || gkey.numeric
+		   || gkey.general_numeric))
     insertkey (&gkey);
   reverse = gkey.reverse;
 
+  if (temp_dir_count == 0)
+    {
+      char const *tmp_dir = getenv ("TMPDIR");
+      add_temp_dir (tmp_dir ? tmp_dir : DEFAULT_TMPDIR);
+    }
+
   if (nfiles == 0)
     {
       nfiles = 1;
@@ -2266,9 +2243,14 @@ but lacks following character offset"));
 
   if (checkonly)
     {
+      if (nfiles > 1)
+	error (SORT_FAILURE, 0,
+	       _("too many arguments;  with -c, there may be at most\
+ one file argument"));
+
       /* POSIX requires that sort return 1 IFF invoked with -c and the
 	 input is not properly sorted.  */
-      exit (check (files, nfiles) == 0 ? 0 : 1);
+      exit (check (files, nfiles) == 0 ? EXIT_SUCCESS : SORT_OUT_OF_ORDER);
     }
 
   if (!STREQ (outfile, "-"))
@@ -2276,6 +2258,7 @@ but lacks following character offset"));
       struct stat outstat;
       if (stat (outfile, &outstat) == 0)
 	{
+	  /* FIXME: warn about this */
 	  /* The following code prevents a race condition when
 	     people use the brain dead shell programming idiom:
 		  cat file | sort -o file
@@ -2372,8 +2355,6 @@ but lacks following character offset"));
 
   if (have_read_stdin && fclose (stdin) == EOF)
     error (SORT_FAILURE, errno, "%s", outfile);
-  if (ferror (stdout) || fclose (stdout) == EOF)
-    error (SORT_FAILURE, errno, _("%s: write error"), outfile);
 
   exit (EXIT_SUCCESS);
 }
